@@ -63,6 +63,7 @@ module Crypto.Store.CMS.Algorithms
     , transportEncrypt
     , transportDecrypt
     , KeyAgreementParams(..)
+    , KeyAgreementKDF(..)
     , ECDHPair
     , ecdhGenerate
     , ecdhPublic
@@ -109,9 +110,11 @@ import           Crypto.Data.Padding
 import           Crypto.ECC (Curve_X25519, Curve_X448, ecdh)
 import           Crypto.Error
 import qualified Crypto.Hash as Hash
+import qualified Crypto.KDF.HKDF as HKDF
 import qualified Crypto.KDF.PBKDF2 as PBKDF2
 import qualified Crypto.KDF.Scrypt as Scrypt
 import qualified Crypto.MAC.HMAC as HMAC
+import qualified Crypto.MAC.KMAC as KMAC
 import qualified Crypto.MAC.Poly1305 as Poly1305
 import           Crypto.Number.Serialize
 import qualified Crypto.PubKey.Curve25519 as X25519
@@ -173,7 +176,27 @@ data DigestProxy hashAlg where
     -- | SHAKE256 (variable size)
     SHAKE256 :: KnownNat n => Proxy n -> DigestProxy (Hash.SHAKE256 n)
 
-deriving instance Show (DigestProxy hashAlg)
+
+instance Show (DigestProxy hashAlg) where
+    showsPrec _ MD2          = showString "MD2"
+    showsPrec _ MD4          = showString "MD4"
+    showsPrec _ MD5          = showString "MD5"
+    showsPrec _ SHA1         = showString "SHA1"
+    showsPrec _ SHA224       = showString "SHA224"
+    showsPrec _ SHA256       = showString "SHA256"
+    showsPrec _ SHA384       = showString "SHA384"
+    showsPrec _ SHA512       = showString "SHA512"
+    showsPrec _ SHAKE128_256 = showString "SHAKE128_256"
+    showsPrec _ SHAKE256_512 = showString "SHAKE256_512"
+    showsPrec d (SHAKE128 p) =
+        showParen (d > 10) $ showString "SHAKE128 " . showNat 11 p
+    showsPrec d (SHAKE256 p) =
+        showParen (d > 10) $ showString "SHAKE256 " . showNat 11 p
+
+showNat :: KnownNat n => Int -> Proxy n -> ShowS
+showNat d n = showParen (d > 0) $
+    shows n . showString " :: Proxy " . shows (natVal n)
+
 deriving instance Eq (DigestProxy hashAlg)
 
 instance HasStrength (DigestProxy hashAlg) where
@@ -300,9 +323,9 @@ instance AlgorithmId DigestAlgorithm where
     parseParameter Type_SHAKE128_256 = parseDigestParam (DigestAlgorithm SHAKE128_256)
     parseParameter Type_SHAKE256_512 = parseDigestParam (DigestAlgorithm SHAKE256_512)
     parseParameter Type_SHAKE128_Len = parseBitLen $
-        \(SomeNat p) -> DigestAlgorithm (SHAKE128 p)
+        \(SomeNat p) -> return $ DigestAlgorithm (SHAKE128 p)
     parseParameter Type_SHAKE256_Len = parseBitLen $
-        \(SomeNat p) -> DigestAlgorithm (SHAKE256 p)
+        \(SomeNat p) -> return $ DigestAlgorithm (SHAKE256 p)
 
 -- | Compute the digest of a message.
 digest :: ByteArrayAccess message => DigestAlgorithm -> message -> ByteString
@@ -318,12 +341,15 @@ hashFromProxy _ = undefined
 parseDigestParam :: Monoid e => DigestAlgorithm -> ParseASN1 e DigestAlgorithm
 parseDigestParam p = getNextMaybe nullOrNothing >> return p
 
-parseBitLen :: Monoid e => (SomeNat -> a) -> ParseASN1 e a
-parseBitLen build = do
+parseBitLen :: Monoid e => (SomeNat -> ParseASN1 e a) -> ParseASN1 e a
+parseBitLen fn = do
     IntVal n <- getNext
     case someNatVal n of
         Nothing -> throwParseError ("Invalid bit length: " ++ show n)
-        Just sn -> return (build sn)
+        Just sn -> fn sn
+
+p256 :: Proxy 256
+p256 = Proxy
 
 p512 :: Proxy 512
 p512 = Proxy
@@ -384,51 +410,114 @@ generateKey params = getRandomBytes (getMaximumKeySize params)
 -- | Message authentication code.  Equality is time constant.
 type MessageAuthenticationCode = AuthTag
 
+data MACType
+    = forall hashAlg . Hash.HashAlgorithm hashAlg
+        => TypeHMAC (DigestProxy hashAlg)
+    | TypeKMAC_SHAKE128
+    | TypeKMAC_SHAKE256
+
+deriving instance Show MACType
+
 -- | Message Authentication Code (MAC) Algorithm.
 data MACAlgorithm
     = forall hashAlg . Hash.HashAlgorithm hashAlg
         => HMAC (DigestProxy hashAlg)
+    | forall n . KnownNat n => KMAC_SHAKE128 (Proxy n) ByteString
+    | forall n . KnownNat n => KMAC_SHAKE256 (Proxy n) ByteString
 
-deriving instance Show MACAlgorithm
+instance Show MACAlgorithm where
+    showsPrec d (HMAC p) = showParen (d > 10) $
+        showString "HMAC " . showsPrec 11 p
+    showsPrec d (KMAC_SHAKE128 p s) = showParen (d > 10) $
+        showString "KMAC_SHAKE128 " . showNat 11 p .
+            showChar ' ' . showsPrec 11 s
+    showsPrec d (KMAC_SHAKE256 p s) = showParen (d > 10) $
+        showString "KMAC_SHAKE256 " . showNat 11 p .
+            showChar ' ' . showsPrec 11 s
 
 instance Eq MACAlgorithm where
-    HMAC a1 == HMAC a2 = DigestAlgorithm a1 == DigestAlgorithm a2
+    HMAC a1             == HMAC a2             =
+        DigestAlgorithm a1 == DigestAlgorithm a2
+    KMAC_SHAKE128 p1 s1 == KMAC_SHAKE128 p2 s2 =
+        natVal p1 == natVal p2 && s1 == s2
+    KMAC_SHAKE256 p1 s1 == KMAC_SHAKE256 p2 s2 =
+        natVal p1 == natVal p2 && s1 == s2
+    _                   == _                   = False
 
 instance HasStrength MACAlgorithm where
     getSecurityBits (HMAC a) = getSecurityBits (DigestAlgorithm a)
+    getSecurityBits (KMAC_SHAKE128 p _) = shakeSecurityBits 128 p
+    getSecurityBits (KMAC_SHAKE256 p _) = shakeSecurityBits 256 p
 
-instance Enumerable MACAlgorithm where
-    values = [ HMAC MD5
-             , HMAC SHA1
-             , HMAC SHA224
-             , HMAC SHA256
-             , HMAC SHA384
-             , HMAC SHA512
+instance Enumerable MACType where
+    values = [ TypeHMAC MD5
+             , TypeHMAC SHA1
+             , TypeHMAC SHA224
+             , TypeHMAC SHA256
+             , TypeHMAC SHA384
+             , TypeHMAC SHA512
+             , TypeKMAC_SHAKE128
+             , TypeKMAC_SHAKE256
              ]
 
-instance OIDable MACAlgorithm where
-    getObjectID (HMAC MD5)    = [1,3,6,1,5,5,8,1,1]
-    getObjectID (HMAC SHA1)   = [1,3,6,1,5,5,8,1,2]
-    getObjectID (HMAC SHA224) = [1,2,840,113549,2,8]
-    getObjectID (HMAC SHA256) = [1,2,840,113549,2,9]
-    getObjectID (HMAC SHA384) = [1,2,840,113549,2,10]
-    getObjectID (HMAC SHA512) = [1,2,840,113549,2,11]
+instance OIDable MACType where
+    getObjectID (TypeHMAC MD5)    = [1,3,6,1,5,5,8,1,1]
+    getObjectID (TypeHMAC SHA1)   = [1,3,6,1,5,5,8,1,2]
+    getObjectID (TypeHMAC SHA224) = [1,2,840,113549,2,8]
+    getObjectID (TypeHMAC SHA256) = [1,2,840,113549,2,9]
+    getObjectID (TypeHMAC SHA384) = [1,2,840,113549,2,10]
+    getObjectID (TypeHMAC SHA512) = [1,2,840,113549,2,11]
+    getObjectID TypeKMAC_SHAKE128 = [2,16,840,1,101,3,4,2,19]
+    getObjectID TypeKMAC_SHAKE256 = [2,16,840,1,101,3,4,2,20]
 
     getObjectID ty = error ("Unsupported MACAlgorithm: " ++ show ty)
 
-instance OIDNameable MACAlgorithm where
+instance OIDNameable MACType where
     fromObjectID oid = unOIDNW <$> fromObjectID oid
 
 instance AlgorithmId MACAlgorithm where
-    type AlgorithmType MACAlgorithm = MACAlgorithm
+    type AlgorithmType MACAlgorithm = MACType
     algorithmName _  = "mac algorithm"
-    algorithmType    = id
-    parameterASN1S _ = id
-    parseParameter p = getNextMaybe nullOrNothing >> return p
+
+    algorithmType (HMAC alg)          = TypeHMAC alg
+    algorithmType (KMAC_SHAKE128 _ _) = TypeKMAC_SHAKE128
+    algorithmType (KMAC_SHAKE256 _ _) = TypeKMAC_SHAKE256
+
+    parameterASN1S (HMAC _)              = id
+    parameterASN1S (KMAC_SHAKE128 p str) = kmacASN1S 256 p str
+    parameterASN1S (KMAC_SHAKE256 p str) = kmacASN1S 512 p str
+
+    parseParameter (TypeHMAC alg)    = getNextMaybe nullOrNothing >> return (HMAC alg)
+    parseParameter TypeKMAC_SHAKE128 = parseKMAC p256 KMAC_SHAKE128
+    parseParameter TypeKMAC_SHAKE256 = parseKMAC p512 KMAC_SHAKE256
+
+kmacASN1S :: (KnownNat n, ASN1Elem e)
+          => Integer -> proxy n -> ByteString -> ASN1Stream e
+kmacASN1S n p str
+    | B.null str && n == i = id
+    | otherwise = asn1Container Sequence (gIntVal i . gOctetString str)
+  where i = natVal p
+
+parseKMAC :: (Monoid e, KnownNat n')
+          => Proxy n'
+          -> (forall n . KnownNat n => Proxy n -> ByteString -> MACAlgorithm)
+          -> ParseASN1 e MACAlgorithm
+parseKMAC p' fn = do
+    b <- hasNext
+    if b then parseParams else return (fn p' B.empty)
+  where
+    parseParams = onNextContainer Sequence $ parseBitLen $ \(SomeNat p) ->
+        getNext >>= \(OctetString str) -> return (fn p str)
 
 instance HasKeySize MACAlgorithm where
     getKeySizeSpecifier (HMAC a) = KeySizeFixed (digestSizeFromProxy a)
-      where digestSizeFromProxy = Hash.hashDigestSize . hashFromProxy
+    getKeySizeSpecifier (KMAC_SHAKE128 p _) =
+        KeySizeFixed (digestSizeFromProxy (SHAKE128 p))
+    getKeySizeSpecifier (KMAC_SHAKE256 p _) =
+        KeySizeFixed (digestSizeFromProxy (SHAKE256 p))
+
+digestSizeFromProxy :: Hash.HashAlgorithm a => proxy a -> Int
+digestSizeFromProxy = Hash.hashDigestSize . hashFromProxy
 
 -- | Invoke the MAC function.
 mac :: (ByteArrayAccess key, ByteArrayAccess message)
@@ -441,6 +530,18 @@ mac (HMAC alg) = hmacWith alg
         => proxy a -> k -> m -> HMAC.HMAC a
     runHMAC _ = HMAC.hmac
 
+mac (KMAC_SHAKE128 p str) = \key -> AuthTag . B.convert . run128 p str key
+  where
+    run128 :: (KnownNat n, ByteArrayAccess k, ByteArrayAccess m)
+           => proxy n -> ByteString -> k -> m -> KMAC.KMAC (Hash.SHAKE128 n)
+    run128 _ = KMAC.kmac
+
+mac (KMAC_SHAKE256 p str) = \key -> AuthTag . B.convert . run256 p str key
+  where
+    run256 :: (KnownNat n, ByteArrayAccess k, ByteArrayAccess m)
+           => proxy n -> ByteString -> k -> m -> KMAC.KMAC (Hash.SHAKE256 n)
+    run256 _ = KMAC.kmac
+
 
 -- Content encryption
 
@@ -1554,36 +1655,60 @@ transportDecrypt _ _ _ = return $ Left UnexpectedPrivateKeyType
 
 -- Key agreement
 
-data KeyAgreementType = TypeStdDH DigestAlgorithm
-                      | TypeCofactorDH DigestAlgorithm
+-- | Key derivation function used for key agreement.
+data KeyAgreementKDF
+    = forall hashAlg . Hash.HashAlgorithm hashAlg
+      => KA_X963_KDF (DigestProxy hashAlg)
+      -- ^ ANSI-X9.63-KDF key derivation function
+    | forall hashAlg . Hash.HashAlgorithm hashAlg
+       => KA_HKDF (DigestProxy hashAlg)
+      -- ^ Extract-and-Expand HMAC-based key derivation function
+
+deriving instance Show KeyAgreementKDF
+
+instance Eq KeyAgreementKDF where
+    KA_X963_KDF a == KA_X963_KDF b = DigestAlgorithm a == DigestAlgorithm b
+    KA_HKDF a     == KA_HKDF b     = DigestAlgorithm a == DigestAlgorithm b
+    _             == _             = False
+
+data KeyAgreementType = TypeStdDH KeyAgreementKDF
+                      | TypeCofactorDH KeyAgreementKDF
                       deriving (Show,Eq)
 
 instance Enumerable KeyAgreementType where
-    values = [ TypeStdDH (DigestAlgorithm SHA1)
-             , TypeStdDH (DigestAlgorithm SHA224)
-             , TypeStdDH (DigestAlgorithm SHA256)
-             , TypeStdDH (DigestAlgorithm SHA384)
-             , TypeStdDH (DigestAlgorithm SHA512)
+    values = [ TypeStdDH (KA_X963_KDF SHA1)
+             , TypeStdDH (KA_X963_KDF SHA224)
+             , TypeStdDH (KA_X963_KDF SHA256)
+             , TypeStdDH (KA_X963_KDF SHA384)
+             , TypeStdDH (KA_X963_KDF SHA512)
 
-             , TypeCofactorDH (DigestAlgorithm SHA1)
-             , TypeCofactorDH (DigestAlgorithm SHA224)
-             , TypeCofactorDH (DigestAlgorithm SHA256)
-             , TypeCofactorDH (DigestAlgorithm SHA384)
-             , TypeCofactorDH (DigestAlgorithm SHA512)
+             , TypeCofactorDH (KA_X963_KDF SHA1)
+             , TypeCofactorDH (KA_X963_KDF SHA224)
+             , TypeCofactorDH (KA_X963_KDF SHA256)
+             , TypeCofactorDH (KA_X963_KDF SHA384)
+             , TypeCofactorDH (KA_X963_KDF SHA512)
+
+             , TypeStdDH (KA_HKDF SHA256)
+             , TypeStdDH (KA_HKDF SHA384)
+             , TypeStdDH (KA_HKDF SHA512)
              ]
 
 instance OIDable KeyAgreementType where
-    getObjectID (TypeStdDH (DigestAlgorithm SHA1))        = [1,3,133,16,840,63,0,2]
-    getObjectID (TypeStdDH (DigestAlgorithm SHA224))      = [1,3,132,1,11,0]
-    getObjectID (TypeStdDH (DigestAlgorithm SHA256))      = [1,3,132,1,11,1]
-    getObjectID (TypeStdDH (DigestAlgorithm SHA384))      = [1,3,132,1,11,2]
-    getObjectID (TypeStdDH (DigestAlgorithm SHA512))      = [1,3,132,1,11,3]
+    getObjectID (TypeStdDH (KA_X963_KDF SHA1))        = [1,3,133,16,840,63,0,2]
+    getObjectID (TypeStdDH (KA_X963_KDF SHA224))      = [1,3,132,1,11,0]
+    getObjectID (TypeStdDH (KA_X963_KDF SHA256))      = [1,3,132,1,11,1]
+    getObjectID (TypeStdDH (KA_X963_KDF SHA384))      = [1,3,132,1,11,2]
+    getObjectID (TypeStdDH (KA_X963_KDF SHA512))      = [1,3,132,1,11,3]
 
-    getObjectID (TypeCofactorDH (DigestAlgorithm SHA1))   = [1,3,133,16,840,63,0,3]
-    getObjectID (TypeCofactorDH (DigestAlgorithm SHA224)) = [1,3,132,1,14,0]
-    getObjectID (TypeCofactorDH (DigestAlgorithm SHA256)) = [1,3,132,1,14,1]
-    getObjectID (TypeCofactorDH (DigestAlgorithm SHA384)) = [1,3,132,1,14,2]
-    getObjectID (TypeCofactorDH (DigestAlgorithm SHA512)) = [1,3,132,1,14,3]
+    getObjectID (TypeCofactorDH (KA_X963_KDF SHA1))   = [1,3,133,16,840,63,0,3]
+    getObjectID (TypeCofactorDH (KA_X963_KDF SHA224)) = [1,3,132,1,14,0]
+    getObjectID (TypeCofactorDH (KA_X963_KDF SHA256)) = [1,3,132,1,14,1]
+    getObjectID (TypeCofactorDH (KA_X963_KDF SHA384)) = [1,3,132,1,14,2]
+    getObjectID (TypeCofactorDH (KA_X963_KDF SHA512)) = [1,3,132,1,14,3]
+
+    getObjectID (TypeStdDH (KA_HKDF SHA256))  = [1,2,840,113549,1,9,16,3,19]
+    getObjectID (TypeStdDH (KA_HKDF SHA384))  = [1,2,840,113549,1,9,16,3,20]
+    getObjectID (TypeStdDH (KA_HKDF SHA512))  = [1,2,840,113549,1,9,16,3,21]
 
     getObjectID ty = error ("Unsupported KeyAgreementType: " ++ show ty)
 
@@ -1591,9 +1716,9 @@ instance OIDNameable KeyAgreementType where
     fromObjectID oid = unOIDNW <$> fromObjectID oid
 
 -- | Key agreement algorithm with associated parameters.
-data KeyAgreementParams = StdDH DigestAlgorithm KeyEncryptionParams
+data KeyAgreementParams = StdDH KeyAgreementKDF KeyEncryptionParams
                           -- ^ 1-Pass D-H with Stardard ECDH
-                        | CofactorDH DigestAlgorithm KeyEncryptionParams
+                        | CofactorDH KeyAgreementKDF KeyEncryptionParams
                           -- ^ 1-Pass D-H with Cofactor ECDH
                         deriving (Show,Eq)
 
@@ -1611,21 +1736,31 @@ instance AlgorithmId KeyAgreementParams where
     parseParameter (TypeCofactorDH d) = CofactorDH d <$> parseAlgorithm Sequence
 
 ecdhKeyMaterial :: (ByteArrayAccess bin, ByteArray bout)
-                => DigestAlgorithm -> KeyEncryptionParams -> Maybe ByteString -> bin -> bout
-ecdhKeyMaterial (DigestAlgorithm hashAlg) kep ukm zz
-    | r == 0    = B.concat (map chunk [1..d])
-    | otherwise = B.concat (map chunk [1..d]) `B.append` B.take r (chunk $ succ d)
-  where
-    (d, r)   = outLen `divMod` Hash.hashDigestSize prx
+                => KeyAgreementKDF -> KeyEncryptionParams -> Maybe ByteString -> bin -> bout
+ecdhKeyMaterial kdf kep ukm zz =
+    case kdf of
+        KA_HKDF hashAlg ->
+            HKDF.expand (extract hashAlg zz) otherInfo outLen
+        KA_X963_KDF hashAlg
+            | r == 0    -> B.concat (map chunk [1..d])
+            | otherwise -> B.concat (map chunk [1..d]) `B.append` B.take r (chunk $ succ d)
+          where
+            (d, r)   = outLen `divMod` Hash.hashDigestSize prx
+            prx      = hashFromProxy hashAlg
 
-    prx      = hashFromProxy hashAlg
+            chunk     = B.convert . Hash.hashFinalize . hashCtx
+            hashCtx'  = Hash.hashUpdate (Hash.hashInitWith prx) zz
+            hashCtx i = Hash.hashUpdate (Hash.hashUpdate hashCtx' $ toWord32 i) otherInfo
+
+  where
     outLen   = getMaximumKeySize kep
     outBits  = 8 * outLen
     toWord32 = i2ospOf_ 4 . fromIntegral
 
-    chunk     = B.convert . Hash.hashFinalize . hashCtx
-    hashCtx'  = Hash.hashInitWith prx
-    hashCtx i = Hash.hashUpdate (Hash.hashUpdate (Hash.hashUpdate hashCtx' zz) (toWord32 i)) otherInfo
+    extract :: (Hash.HashAlgorithm a, ByteArrayAccess ikm)
+            => DigestProxy a -> ikm -> HKDF.PRK a
+    extract _ = HKDF.extract (fromMaybe B.empty ukm)
+
     otherInfo =
         let ki  = algorithmASN1S Sequence kep
             eui = case ukm of
